@@ -23,6 +23,7 @@ from optax._src import base
 from optax._src import clipping
 from optax._src import combine
 from optax._src import factorized
+from optax._src import linesearch as _linesearch
 from optax._src import transform
 from optax._src import wrappers
 
@@ -36,15 +37,43 @@ def adabelief(
     b2: float = 0.999,
     eps: float = 1e-16,
     eps_root: float = 1e-16) -> base.GradientTransformation:
-  """The AdaBelief optimizer.
+  r"""The AdaBelief optimizer.
 
   AdaBelief is an adaptive learning rate optimizer that focuses on fast
   convergence, generalization, and stability. It adapts the step size depending
   on its "belief" in the gradient direction — the optimizer adaptively scales
   the step size by the difference between the predicted and observed gradients.
-  AdaBelief is a modified version of Adam and contains the same number of
-  parameters.
-  
+  AdaBelief is a modified version of :func:`optax.adam` and contains the same
+  number of parameters.
+
+  Let :math:`\alpha_t` represent the learning rate and :math:`\beta_1, \beta_2`,
+  :math:`\varepsilon`, :math:`\bar{\varepsilon}` represent the arguments
+  ``b1``, ``b2``, ``eps`` and ``eps_root`` respectively. The learning rate is
+  indexed by :math:`t` since the learning rate may also be provided by a
+  schedule function.
+
+  The ``init`` function of this optimizer initializes an internal state
+  :math:`S_0 := (m_0, s_0) = (0, 0)`, representing initial estimates for the
+  first and second moments. In practice these values are stored as pytrees
+  containing all zeros, with the same shape as the model updates.
+  At step :math:`t`, the ``update`` function of this optimizer takes as
+  arguments the incoming gradients :math:`g_t` and optimizer state :math:`S_t`
+  and computes updates :math:`u_t` and new state :math:`S_{t+1}`. Thus, for
+  :math:`t > 0`, we have,
+
+  .. math::
+
+    \begin{align*}
+      m_t &\leftarrow \beta_1 \cdot m_{t-1} + (1-\beta_1) \cdot g_t \\
+      s_t &\leftarrow \beta_2 \cdot s_{t-1} + (1-\beta_2) \cdot (g_t - m_t)^2 
+      + \bar{\varepsilon} \\
+      \hat{m}_t &\leftarrow m_t / {(1-\beta_1^t)} \\
+      \hat{s}_t &\leftarrow s_t / {(1-\beta_2^t)} \\
+      u_t &\leftarrow -\alpha_t \cdot \hat{m}_t / \left(\sqrt{\hat{s}_t}
+      + \varepsilon \right) \\
+      S_t &\leftarrow (m_t, s_t).
+    \end{align*}
+
   Examples:
     >>> import optax
     >>> import jax
@@ -322,7 +351,7 @@ def adam(
 
   Let :math:`\alpha_t` represent the learning rate and :math:`\beta_1, \beta_2`,
   :math:`\varepsilon`, :math:`\bar{\varepsilon}` represent the arguments
-  ``b1``, ``b2``, ``eps`` and ``eps_root`` respectievly. The learning rate is
+  ``b1``, ``b2``, ``eps`` and ``eps_root`` respectively. The learning rate is
   indexed by :math:`t` since the learning rate may also be provided by a
   schedule function.
 
@@ -502,15 +531,50 @@ def adamw(
     *,
     nesterov: bool = False,
 ) -> base.GradientTransformation:
-  """Adam with weight decay regularization.
+  r"""Adam with weight decay regularization.
 
   AdamW uses weight decay to regularize learning towards small weights, as
   this leads to better generalization. In SGD you can also use L2 regularization
   to implement this as an additive loss term, however L2 regularization
-  does not behave as intended for adaptive gradient algorithms such as Adam.
+  does not behave as intended for adaptive gradient algorithms such as Adam,
+  see [Loshchilov et al, 2019].
+
+  Let :math:`\alpha_t` represent the learning rate and :math:`\beta_1, \beta_2`,
+  :math:`\varepsilon`, :math:`\bar{\varepsilon}` represent the arguments
+  ``b1``, ``b2``, ``eps`` and ``eps_root`` respectively. The learning rate is
+  indexed by :math:`t` since the learning rate may also be provided by a
+  schedule function. Let :math:`\lambda` be the weight decay and 
+  :math:`\theta_t` the parameter vector at time :math:`t`.
+
+  The ``init`` function of this optimizer initializes an internal state
+  :math:`S_0 := (m_0, v_0) = (0, 0)`, representing initial estimates for the
+  first and second moments. In practice these values are stored as pytrees
+  containing all zeros, with the same shape as the model updates.
+  At step :math:`t`, the ``update`` function of this optimizer takes as
+  arguments the incoming gradients :math:`g_t`, the optimizer state :math:`S_t` 
+  and the parameters :math:`\theta_t` and computes updates :math:`u_t` and 
+  new state :math:`S_{t+1}`. Thus, for :math:`t > 0`, we have,
+
+  .. math::
+
+    \begin{align*}
+      m_t &\leftarrow \beta_1 \cdot m_{t-1} + (1-\beta_1) \cdot g_t \\
+      v_t &\leftarrow \beta_2 \cdot v_{t-1} + (1-\beta_2) \cdot {g_t}^2 \\
+      \hat{m}_t &\leftarrow m_t / {(1-\beta_1^t)} \\
+      \hat{v}_t &\leftarrow v_t / {(1-\beta_2^t)} \\
+      u_t &\leftarrow -\alpha_t \cdot \left( \hat{m}_t / \left({\sqrt{\hat{v}_t 
+      + \bar{\varepsilon}} + \varepsilon} \right) + \lambda \theta_{t} \right)\\
+      S_t &\leftarrow (m_t, v_t).
+    \end{align*}
 
   This implementation can incorporate a momentum a la Nesterov introduced by
   [Dozat 2016]. The resulting optimizer is then often referred as NAdamW.
+  With the keyword argument `nesterov=True`, the optimizer uses Nesterov
+  momentum, replacing the above :math:`\hat{m}_t` with
+
+  .. math::
+      \hat{m}_t \leftarrow
+        \beta_1 m_t / {(1-\beta_1^{t+1})} + (1 - \beta_1) g_t / {(1-\beta_1^t)}.
   
   Examples:
     >>> import optax
@@ -994,9 +1058,27 @@ def noisy_sgd(
 ) -> base.GradientTransformation:
   r"""A variant of SGD with added noise.
 
-  It has been found that adding noise to the gradients can improve
-  both the training error and the generalization error in very deep networks.
-  
+  Noisy SGD is a variant of :func:`optax.sgd` that incorporates Gaussian noise
+  into the updates. It has been found that adding noise to the gradients can
+  improve both the training error and the generalization error in very deep
+  networks.
+
+  The update :math:`u_t` is modified to include this noise as follows:
+
+  .. math::
+    u_t \leftarrow -\alpha_t (g_t + N(0, \sigma_t^2)),
+
+  where :math:`N(0, \sigma_t^2)` represents Gaussian noise with zero mean and a
+  variance of :math:`\sigma_t^2`.
+
+  The variance of this noise decays over time according to the formula
+
+  .. math::
+    \sigma_t^2 = \frac{\eta}{(1+t)^\gamma},
+
+  where :math:`\gamma` is the decay rate parameter ``gamma`` and :math:`\eta`
+  represents the initial variance ``eta``.
+
   Examples:
     >>> import optax
     >>> import jax
@@ -1025,8 +1107,8 @@ def noisy_sgd(
     learning_rate: A global scaling factor, either fixed or evolving along
       iterations with a scheduler, see :func:`optax.scale_by_learning_rate`.
     eta: Initial variance for the Gaussian noise added to gradients.
-    gamma: A parameter controlling the annealing of noise over time, the
-      variance decays according to `(1+t)^-\gamma`.
+    gamma: A parameter controlling the annealing of noise over time ``t``, the
+      variance decays according to ``(1+t)**(-gamma)``.
     seed: Seed for the pseudo-random generation process.
 
   Returns:
@@ -1160,7 +1242,9 @@ def radam(
     b2: float = 0.999,
     eps: float = 1e-8,
     eps_root: float = 0.0,
-    threshold: float = 5.0
+    threshold: float = 5.0,
+    *,
+    nesterov: bool = False,
 ) -> base.GradientTransformation:
   """The Rectified Adam optimizer.
 
@@ -1204,13 +1288,20 @@ def radam(
       in RMSProp), to avoid dividing by zero when rescaling. This is needed for
       instance when computing (meta-)gradients through Adam.
     threshold: Threshold for variance tractability.
+    nesterov: Whether to use Nesterov momentum.
 
   Returns:
     The corresponding `GradientTransformation`.
   """
   return combine.chain(
       transform.scale_by_radam(
-          b1=b1, b2=b2, eps=eps, eps_root=eps_root, threshold=threshold),
+          b1=b1,
+          b2=b2,
+          eps=eps,
+          eps_root=eps_root,
+          threshold=threshold,
+          nesterov=nesterov,
+      ),
       transform.scale_by_learning_rate(learning_rate),
   )
 
@@ -1327,12 +1418,12 @@ def sgd(
   .. math::
 
     \begin{align*}
-      m_t & \leftarrow \begin{cases}
-        g_t + \mu m_t & \mathrm{if \ \texttt{nesterov=False}}\\
-        g_t + \mu (g_t + \mu m_{t-1}) & \mathrm{if \ \texttt{nesterov=True}}
-      \end{cases} \\
-      u_t & \leftarrow - \alpha_t m_t \\
-      S_t & = m_t,
+      m_t &\leftarrow g_t + \mu m_{t-1} \\
+      u_t &\leftarrow \begin{cases}
+        -\alpha_t m_t & \text{ if } \texttt{nesterov = False} \\
+        -\alpha_t (g_t + \mu m_t) & \text{ if } \texttt{nesterov = True}
+        \end{cases} \\
+      S_t &\leftarrow m_t,
     \end{align*}
 
   where :math:`\mu` is the ``momentum`` parameter and :math:`S_t` is the state
@@ -1500,7 +1591,37 @@ def adamax(
     b2: float = 0.999,
     eps: float = 1e-8,
 ) -> base.GradientTransformation:
-  """A variant of the Adam optimizer that uses the infinity norm.
+  r"""A variant of the Adam optimizer that uses the infinity norm.
+  
+  AdaMax is a variant of the :func:`optax.adam` optimizer. By generalizing 
+  Adam's :math:`L^2` norm to an :math:`L^p` norm and taking the limit as
+  :math:`p \rightarrow \infty`, we obtain a simple and stable update rule.
+
+  Let :math:`\alpha_t` represent the learning rate and :math:`\beta_1, \beta_2`,
+  :math:`\varepsilon` represent the arguments
+  ``b1``, ``b2`` and ``eps`` respectively. The learning rate is
+  indexed by :math:`t` since the learning rate may also be provided by a
+  schedule function.
+
+  The ``init`` function of this optimizer initializes an internal state
+  :math:`S_0 := (m_0, v_0) = (0, 0)`, representing initial estimates for the
+  first and second moments. In practice these values are stored as pytrees
+  containing all zeros, with the same shape as the model updates.
+  At step :math:`t`, the ``update`` function of this optimizer takes as
+  arguments the incoming gradients :math:`g_t` and optimizer state :math:`S_t`
+  and computes updates :math:`u_t` and new state :math:`S_{t+1}`. Thus, for
+  :math:`t > 0`, we have,
+
+  .. math::
+
+    \begin{align*}
+      m_t &\leftarrow \beta_1 \cdot m_{t-1} + (1-\beta_1) \cdot g_t \\
+      v_t &\leftarrow \max(\left| g_t \right| + \varepsilon, \beta_2 \cdot
+      v_{t-1}) \\
+      \hat{m}_t &\leftarrow m_t / (1-\beta_1^t) \\
+      u_t &\leftarrow -\alpha_t \cdot \hat{m}_t / v_t \\
+      S_t &\leftarrow (m_t, v_t).
+    \end{align*}
   
   Examples:
     >>> import optax
@@ -1631,7 +1752,7 @@ def rprop(
   gradient descent. It responds only to the sign of the gradient by increasing
   or decreasing the step size selected per parameter exponentially to speed up
   convergence and avoid oscillations.
-  
+
   Examples:
     >>> import optax
     >>> import jax
@@ -1654,11 +1775,14 @@ def rprop(
     Objective function: 1.38E+01
 
   References:
-    PyTorch implementation:
-      https://pytorch.org/docs/stable/generated/torch.optim.Rprop.html
-    Riedmiller and Braun, 1993: https://ieeexplore.ieee.org/document/298623
-    Igel and Hüsken, 2003:
-      https://www.sciencedirect.com/science/article/abs/pii/S0925231201007007
+    Riedmiller and Braun. `A direct adaptive method for faster backpropagation 
+    learning: the RPROP algorithm 
+    <https://ieeexplore.ieee.org/document/298623>`_, 1993
+
+    Igel and Hüsken.  `Empirical evaluation of the improved Rprop learning 
+    algorithms
+    <https://www.sciencedirect.com/science/article/abs/pii/S0925231201007007>`_,
+    2003
 
   Args:
     learning_rate: The initial step size.
@@ -1684,3 +1808,215 @@ def rprop(
       ),
       transform.scale(-1.0),
   )
+
+
+def polyak_sgd(
+    max_learning_rate: float = 1.,
+    scaling: base.ScalarOrSchedule = 1.,
+    f_min: float = 0.0,
+    eps: float = 0.0,
+) -> base.GradientTransformationExtraArgs:
+  r"""SGD with Polyak step-size.
+
+  This solver implements the SGD with Polyak step size of (Loizou et al. 2021).
+  It sets the step-size as
+
+  .. math::
+    s \min\left\{\frac{f(x) - f^\star}{\|\nabla f(x)\|^2 + \epsilon},
+      \gamma_{\max}\right\}\,,
+
+  where :math:`f` is the function from which a gradient is computed,
+  :math:`\gamma_{\max}` is a maximal acceptable learning rate set  by
+  ``max_learning_rate``, :math:`\epsilon` is a constant preventing division by
+  zero set with ``eps``, :math:`s` scales the formula by ``scaling``, and
+  :math:`f^\star` is a guess of the minimum value of the function set with
+  ``f_min``.
+
+
+  Examples:
+    >>> import optax
+    >>> import jax
+    >>> import jax.numpy as jnp
+    >>> def f(x): return jnp.sum(x ** 2)  # simple quadratic function
+    >>> solver = optax.polyak_sgd()
+    >>> params = jnp.array([1., 2., 3.])
+    >>> print('Objective function: ', f(params))
+    Objective function:  14.0
+    >>> opt_state = solver.init(params)
+    >>> for _ in range(5):
+    ...  value, grad = jax.value_and_grad(f)(params)
+    ...  params, opt_state = solver.update(grad, opt_state, params, value=value)
+    ...  print('Objective function: ', f(params))
+    Objective function:  3.5
+    Objective function:  0.875
+    Objective function:  0.21875
+    Objective function:  0.0546875
+    Objective function:  0.013671875
+
+  .. warning::
+    This method requires knowledge of an approximate value of the of the
+    objective function minimum, passed through the ``f_min`` argument.
+    For models that interpolate the data, this can be set to 0 (default
+    value).
+    Failing to set an appropriate value for ``f_min`` can lead to
+    divergence or convergence to a suboptimal solution.
+
+  References:
+    Loizou et al. `Stochastic polyak step-size for SGD: An adaptive learning
+    rate for fast convergence <https://arxiv.org/abs/2002.10542>`_, 2021
+
+    Berrada et al., `Training neural networks for and by interpolation
+    <https://arxiv.org/pdf/1906.05661.pdf>`_, 2020
+
+  Args:
+    max_learning_rate: a maximum step size to use (defaults to 1).
+    scaling: A global scaling factor, either fixed or evolving along
+      iterations with a scheduler (defaults to 1).
+    f_min: a lower bound on the objective function (defaults to 0). Corresponds
+      to :math:`f^\star` in the formula above.
+    eps: a value to add in the denominator of the update (defaults to 0).
+
+  Returns:
+    A :class:`GradientTransformationExtraArgs`, where the ``update`` function
+    takes an additional keyword argument ``value`` containing the current
+    value of the objective function.
+  """
+  return combine.chain(
+      sgd(learning_rate=scaling),
+      transform.scale_by_polyak(
+          max_learning_rate=max_learning_rate, f_min=f_min, eps=eps
+      ),
+  )
+
+
+def lbfgs(
+    learning_rate: Optional[base.ScalarOrSchedule] = None,
+    memory_size: int = 10,
+    scale_init_precond: bool = True,
+    linesearch: Optional[
+        base.GradientTransformationExtraArgs
+    ] = _linesearch.scale_by_zoom_linesearch(max_linesearch_steps=15),
+) -> base.GradientTransformationExtraArgs:
+  r"""L-BFGS optimizer.
+
+  L-BFGS is a quasi-Newton method that multiplies the update (gradient)
+  with an approximation of the inverse Hessian. This algorithm does not need
+  access to the Hessian, as this approximation is constructed from the gradient
+  evaluations seen during optimization. L-BFGS is a limited-memory variant of
+  the Broyden-Fletcher-Goldfarb-Shanno (BFGS) algorithm. The BFGS algorithm
+  requires storing a matrix of size :math:`p \times p` with :math:`p` the
+  dimension of the parameters.
+  The limited variant circuments this issue by computing the approximation of
+  the inverse using only :math:`m` (``memory_size``) past differences of
+  parameters/gradients. Namely, the approximation of the Hessian inverse is
+  denoted :math:`P_k = P_{k, k}`, where
+
+  .. math::
+
+    \begin{align*}
+      P_{k, j+1} & = V_j^\top P_{k, j} V_j + \rho_j \delta w_j \delta w_j^\top
+      \quad \text{for} \ j \in \{k-m, \ldots, k-1\}\\
+      P_{k, k-m} & = \gamma_k I \\
+      V_k & = I - \rho_k \delta u_k \delta w_k^\top \\
+      \rho_k & = 1/(\delta u_k^\top \delta w_k) \\
+      \delta w_k & = w_{k+1} - w_k \\
+      \delta u_k & = u_{k+1} - u_k \\
+      \gamma_k & =
+        \begin{cases}
+          (\delta w_{k-1}^\top \delta u_{k-1}) /
+          (\delta u_{k-1}^\top \delta u_{k-1})
+          & \text{if} \ \texttt{scale\_init\_hess} \\
+          1 & \text{otherwise}
+        \end{cases},
+    \end{align*}
+
+  for
+  :math:`u_k` the gradients/updates at iteration :math:`k`,
+  :math:`w_k` the parameters at iteration :math:`k`.
+
+  The formula for updating :math:`P_k` is obtained by computing the optimal
+  preconditioning matrix subject to some secant condition, see references
+  for more details. Computing :math:`P_k u_k` can be done by a sequence of 
+  vector operations using past differences of parameters and gradients stored in
+  a memory bufffer.
+
+  The present function just outputs the LBFGS direction :math:`P_k u_k`.
+  It can be chained with a linesearch ensuring sufficient decrease and low
+  curvature, such as a zoom linesearch. The linesearch computes a stepsize
+  :math:`\eta_k`, such that the updated parameters
+  (using :func:`optax.apply_updates`) take the form
+  :math:`w_{k+1} = w_k - \eta_k P_k u_k`.
+
+  Example:
+    >>> import optax
+    >>> import jax
+    >>> import jax.numpy as jnp
+    >>> def f(x): return jnp.sum(x ** 2)
+    >>> solver = optax.lbfgs()
+    >>> params = jnp.array([1., 2., 3.])
+    >>> print('Objective function: ', f(params))
+    Objective function:  14.0
+    >>> opt_state = solver.init(params)
+    >>> value_and_grad = optax.value_and_grad_from_state(f)
+    >>> for _ in range(5):
+    ...   value, grad = value_and_grad(params, state=opt_state)
+    ...   updates, opt_state = solver.update(
+    ...      grad, opt_state, params, value=value, grad=grad, value_fn=f
+    ...   )
+    ...   params = optax.apply_updates(params, updates)
+    ...   print('Objective function: ', f(params))
+    Objective function:  0.0
+    Objective function:  0.0
+    Objective function:  0.0
+    Objective function:  0.0
+    Objective function:  0.0
+
+  .. warning::
+    This method is memory intensive optimizer, best used for small to medium
+    scale problems.
+
+  .. warning::
+    This optimizer works best with a linesearch (current default is a
+    zoom linesearch). See example above for best use in a non-stochastic
+    setting, where we can recycle gradients computed by the linesearch using
+    :func:`optax.value_and_grad_from_state`.
+
+  References:
+    Algorithms 7.4, 7.5 (page 199) of Nocedal et al, `Numerical Optimization
+    <https://www.math.uci.edu/~qnie/Publications/NumericalOptimization.pdf>`_
+    , 1999
+
+    Liu et al., `On the limited memory BFGS method for large scale optimization
+    <https://users.iems.northwestern.edu/~nocedal/PDFfiles/limited-memory.pdf>`_
+    , 1989.
+
+  Args:
+    learning_rate: optional global scaling factor, either fixed or evolving
+      along iterations with a scheduler, see 
+      :func:`optax.scale_by_learning_rate`. By default the learning rate is
+      handled by a linesearch.
+    memory_size: number of past updates to keep in memory to approximate the
+      Hessian inverse.
+    scale_init_precond: whether to use a scaled identity as the initial
+      preconditioner, see formula above.
+    linesearch: an instance of :class:`optax.GradientTransformationExtraArgs`
+      such as :func:`optax.scale_by_zoom_linesearch` that computes a
+      learning rate, a.k.a. stepsize, to satisfy some criterion such as a
+      sufficient decrease of the objective by additional calls to the objective.
+
+  Returns:
+    A :class:`optax.GradientTransformationExtraArgs` object.
+  """
+  if learning_rate is None:
+    base_scaling = transform.scale(-1.0)
+  else:
+    base_scaling = transform.scale_by_learning_rate(learning_rate)
+  if linesearch is None:
+    linesearch = base.identity()
+  return combine.chain(
+      transform.scale_by_lbfgs(
+          memory_size=memory_size, scale_init_precond=scale_init_precond
+      ),
+      base_scaling,
+      linesearch,
+  )
